from PySide6.QtWidgets import (
    QCheckBox,
    QLabel,
    QWidget,
    QGridLayout,
    QLineEdit,
    QApplication,
    QTabWidget,
    QPushButton,
    QScrollArea,
    QFrame,
    QRadioButton,
    QButtonGroup,
    QSpacerItem,
    QSizePolicy,
)
from PySide6.QtCore import Signal, QThread
from PySide6.QtGui import QFont

from nomad_camels.main_classes.manual_control import (
    Manual_Control,
    Manual_Control_Config,
)
from nomad_camels.ui_widgets.channels_check_table import Channels_Check_Table
from nomad_camels.utility import device_handling, variables_handling

import time

bold_font = QFont()
bold_font.setBold(True)


class Set_Panel(Manual_Control):
    """This class provides convenience to set given values quickly at the push
    of a button. Further, it provides a utility-instrument using the buttons as
    channels to easily use the functionality in protocols."""

    def __init__(self, parent=None, control_data=None):
        control_data = control_data or {}
        if "name" in control_data:
            name = control_data["name"]
        else:
            name = "Step Panel"
        super().__init__(parent=parent, title=name)
        self.horizontal = (
            control_data["horizontal"] if "horizontal" in control_data else False
        )
        self.setLayout(QGridLayout())
        groups = (
            control_data["button_groups"] if "button_groups" in control_data else {}
        )

        channels = []
        self.buttons = []
        self.set_vals = []
        self.button_groups = []

        for n, (group, group_data) in enumerate(groups.items()):
            self.buttons.append([])
            self.set_vals.append([])
            button_group = QButtonGroup()
            self.button_groups.append(button_group)
            group_widget = QFrame()
            layout = QGridLayout()
            group_widget.setLayout(layout)
            layout.setContentsMargins(0, 0, 0, 0)

            label = QLabel(group)
            label.setFont(bold_font)
            layout.addWidget(label, 0, 0)
            if self.horizontal:
                spacer = QSpacerItem(0, 1, QSizePolicy.Expanding, QSizePolicy.Minimum)
                for i, (button, set_vals) in enumerate(group_data.items()):
                    radio_button = QRadioButton(button)
                    radio_button.clicked.connect(self.button_pushed)
                    layout.addWidget(radio_button, 0, 1 + i)
                    self.buttons[n].append(radio_button)
                    self.set_vals[n].append(set_vals)
                    button_group.addButton(radio_button)
                layout.addItem(spacer, 0, len(group_data) + 1)
                self.layout().addWidget(group_widget, n, 0)
            else:
                spacer = QSpacerItem(1, 0, QSizePolicy.Minimum, QSizePolicy.Expanding)
                for i, (button, set_vals) in enumerate(group_data.items()):
                    radio_button = QRadioButton(button)
                    radio_button.clicked.connect(self.button_pushed)
                    layout.addWidget(radio_button, 1 + i, 0)
                    self.buttons[n].append(radio_button)
                    self.set_vals[n].append(set_vals)
                    button_group.addButton(radio_button)
                layout.addItem(spacer, len(group_data) + 1, 0)
                self.layout().addWidget(group_widget, 0, n)
            for data in group_data.values():
                channels += data["channel"]
        self.channel_list = list(set(channels))
        self.adjustSize()
        self.read_thread = None
        self.channels = None
        self.start_multiple_devices(self.channel_list, True)

    def device_ready(self):
        self.channels = device_handling.get_channels_from_string_list(
            self.channel_list, True
        )
        self.start_read_thread()
        super().device_ready()

    def read_error(self, e):
        """Shows a message box with the error message. And asks wether the read thread should be restarted"""
        from PySide6.QtWidgets import QMessageBox

        msg = QMessageBox()
        msg.setIcon(QMessageBox.Critical)
        msg.setText("Error in read thread")
        msg.setInformativeText(
            f"An error occured in the read thread of {self.name}:\n{e}\nDo you want to restart the read thread?"
        )
        msg.setWindowTitle("Error in read thread")
        msg.setStandardButtons(QMessageBox.Yes | QMessageBox.No)
        msg.setDefaultButton(QMessageBox.Yes)
        if msg.exec() == QMessageBox.Yes:
            self.start_read_thread()

    def start_read_thread(self):
        t = (
            self.control_data["readback_time"]
            if "readback_time" in self.control_data
            else 5
        )
        self.read_thread = Readback_Thread(self, self.channels, t)
        if "readback" in self.control_data and self.control_data["readback"]:
            self.read_thread.data_sig.connect(self.check_readback)
            self.read_thread.exception_signal.connect(self.read_error)
            self.read_thread.start()

    def button_pushed(self):
        for n, group in enumerate(self.buttons):
            for i, button in enumerate(group):
                if button.isChecked():
                    for j, channel in enumerate(self.set_vals[n][i]["channel"]):
                        value = self.set_vals[n][i]["value"][j]
                        value = conv_value(value)
                        self.channels[channel].put(value)

    def check_readback(self, vals):
        for n, group in enumerate(self.buttons):
            self.button_groups[n].setExclusive(False)
            for i, button in enumerate(group):
                is_same = True
                for j, channel in enumerate(self.set_vals[n][i]["channel"]):
                    set_value = self.set_vals[n][i]["value"][j]
                    set_value = conv_value(set_value)
                    is_value = conv_value(str(vals[channel]))
                    if set_value != is_value:
                        is_same = False
                        break
                button.setChecked(is_same)
            self.button_groups[n].setExclusive(True)

    def close(self) -> bool:
        if self.read_thread:
            self.read_thread.still_running = False
        return super().close()

    def closeEvent(self, a0) -> None:
        if self.read_thread:
            self.read_thread.still_running = False
        return super().closeEvent(a0)


def conv_value(value):
    try:
        value = int(value)
    except ValueError:
        try:
            value = float(value)
        except ValueError:
            try:
                value = bool(value)
            except ValueError:
                pass
    return value


class Readback_Thread(QThread):
    data_sig = Signal(dict)
    exception_signal = Signal(Exception)

    def __init__(self, parent=None, channels=None, read_time=5):
        super().__init__(parent=parent)
        self.channels = channels or []
        self.read_time = read_time
        self.still_running = True

    def run(self):
        try:
            self.do_reading()
            accum = 0
            while self.still_running:
                if self.read_time > 5:
                    if self.read_time - accum > 5:
                        time.sleep(5)
                        accum += 5
                        continue
                    else:
                        time.sleep(self.read_time - accum)
                        accum = 0
                else:
                    time.sleep(self.read_time)
                self.do_reading()
        except Exception as e:
            self.exception_signal.emit(e)

    def do_reading(self):
        vals = {}
        for name, channel in self.channels.items():
            vals[name] = channel.get()
        self.data_sig.emit(vals)


class Set_Panel_Config(Manual_Control_Config):
    """ """

    def __init__(self, parent=None, control_data=None):
        super().__init__(
            parent=parent,
            control_data=control_data,
            title="Set Panel Config",
            control_type="Set_Panel",
        )
        control_data = control_data or {}
        label_button_groups = QLabel("Button Groups")
        label_button_groups.setStyleSheet("font-size: 9pt")
        label_button_groups.setFont(bold_font)
        self.checkbox_readback = QCheckBox("readback loop in s:")
        if "readback" in control_data:
            self.checkbox_readback.setChecked(control_data["readback"])
        self.lineEdit_readback = QLineEdit("5")
        if "readback_time" in control_data:
            self.lineEdit_readback.setText(str(control_data["readback_time"]))
        self.checkBox_as_instr = QCheckBox("provide set panel as instrument")
        if "provide_instr" in control_data:
            self.checkBox_as_instr.setChecked(control_data["provide_instr"])
        self.checkBox_groups_horizontal = QCheckBox("arrange groups horizontally")
        if "horizontal" in control_data:
            self.checkBox_groups_horizontal.setChecked(control_data["horizontal"])
        self.pushButton_add_group = QPushButton("add button group")
        self.pushButton_add_group.clicked.connect(self.add_button_group)

        self.tabs = []
        self.tab_widget = QTabWidget()
        if "button_groups" in control_data:
            for name, group in control_data["button_groups"].items():
                self.add_button_group(name, group)
        if not self.tabs:
            self.add_button_group()
        self.change_tab_name()

        layout = self.layout()
        layout.addWidget(self.checkBox_as_instr, 1, 0, 1, 2)
        layout.addWidget(self.checkbox_readback, 2, 0)
        layout.addWidget(self.lineEdit_readback, 2, 1)
        layout.addWidget(label_button_groups, 3, 0)
        layout.addWidget(self.checkBox_groups_horizontal, 3, 1)
        layout.addWidget(self.pushButton_add_group, 4, 0, 1, 2)
        layout.addWidget(self.tab_widget, 6, 0, 1, 2)

    def remove_button_group(self):
        """ """
        ind = self.tab_widget.currentIndex()
        self.tab_widget.removeTab(ind)
        self.tabs.pop(ind)

    def add_button_group(self, name="", group_data=None):
        """

        Parameters
        ----------


        Returns
        -------

        """
        tab = Button_Group_Tab(name, group_data, self)
        tab.signal_name_change.connect(self.change_tab_name)
        tab.signal_remove.connect(self.remove_button_group)
        tab.signal_move_left.connect(lambda: self.move_tab(-1))
        tab.signal_move_right.connect(lambda: self.move_tab(1))
        self.tabs.append(tab)
        self.tab_widget.addTab(tab, "")

    def change_tab_name(self):
        """ """
        for i, tab in enumerate(self.tabs):
            name = tab.lineEdit_group_name.text()
            self.tab_widget.setTabText(i, name)

    def move_tab(self, direction=1):
        """

        Parameters
        ----------
        direction :
             (Default value = 1)

        Returns
        -------

        """
        ind = self.tab_widget.currentIndex()
        self.tab_widget.tabBar().moveTab(ind, ind + direction)
        self.tabs[ind + direction], self.tabs[ind] = (
            self.tabs[ind],
            self.tabs[ind + direction],
        )

    def accept(self):
        self.control_data["horizontal"] = self.checkBox_groups_horizontal.isChecked()
        if "button_groups" not in self.control_data:
            self.control_data["button_groups"] = {}
        button_groups = self.control_data["button_groups"]
        button_groups.clear()
        for group in self.tabs:
            button_groups.update(group.get_data())
        self.control_data["provide_instr"] = self.checkBox_as_instr.isChecked()
        self.control_data["readback"] = self.checkbox_readback.isChecked()
        self.control_data["readback_time"] = float(self.lineEdit_readback.text())
        if self.checkBox_as_instr.isChecked():
            self.provide_instrument()
        else:
            self.remove_instrument()
        super().accept()

    def provide_instrument(self):
        groups = self.control_data["button_groups"]
        set_vals = []
        channels = []
        group_names = []
        for n, (group, group_data) in enumerate(groups.items()):
            set_vals.append([])
            group_names.append(group)
            for i, (button, vals) in enumerate(group_data.items()):
                set_vals[n].append(vals)
            for data in group_data.values():
                channels += data["channel"]
        channels = list(set(channels))
        non_str_channels = {}
        str_channels = {}
        for channel in channels:
            str_channels[channel] = channel
            non_str_channels[channel] = variables_handling.channels[
                channel
            ].get_bluesky_name()
        settings = {
            "set_vals": set_vals,
            "group_names": group_names,
            "channel_names": channels,
            "!non_string!_channels": non_str_channels,
        }
        from .nomad_camels_driver_set_panel_device.set_panel_device import subclass

        dev = subclass(**settings)
        dev.settings = settings
        name = f"{self.lineEdit_name.text()}_manual_control"
        dev.custom_name = name
        dev.get_channels()
        variables_handling.devices[name] = dev
        variables_handling.channels.update(dev.get_channels())

    def remove_instrument(self):
        name = f"{self.lineEdit_name.text()}_manual_control"
        if name in variables_handling.devices:
            dev = variables_handling.devices[name]
            for chan in dev.get_channels():
                variables_handling.channels.pop(chan)
            variables_handling.devices.pop(name)


class Button_Group_Tab(QWidget):
    signal_remove = Signal()
    signal_move_left = Signal()
    signal_move_right = Signal()
    signal_name_change = Signal(str)

    def __init__(self, name="", group_data=None, parent=None):
        super().__init__(parent)
        group_data = group_data or {}
        label_group_name = QLabel("Group Name:")
        self.lineEdit_group_name = QLineEdit(name)
        self.lineEdit_group_name.textChanged.connect(self.signal_name_change.emit)

        self.moveLeftButton = QPushButton("Move left")
        self.moveRightButton = QPushButton("Move right")
        self.removeButton = QPushButton("Remove")

        self.moveRightButton.clicked.connect(self.signal_move_right.emit)
        self.moveLeftButton.clicked.connect(self.signal_move_left.emit)
        self.removeButton.clicked.connect(self.signal_remove.emit)

        self.add_button_button = QPushButton("Add Button")
        self.add_button_button.clicked.connect(self.add_button)
        self.button_widge = QWidget()
        self.button_widge.setLayout(QGridLayout())
        button_area = QScrollArea()
        button_area.setWidget(self.button_widge)
        button_area.setWidgetResizable(True)

        self.buttons = {
            "labels": [],
            "texts": [],
            "tables": [],
            "removers": [],
            "lines": [],
        }

        layout = QGridLayout()
        self.setLayout(layout)
        layout.addWidget(self.moveLeftButton, 2, 0)
        layout.addWidget(self.moveRightButton, 2, 1)
        layout.addWidget(self.removeButton, 2, 2)
        layout.addWidget(label_group_name, 4, 0)
        layout.addWidget(self.lineEdit_group_name, 4, 1, 1, 2)
        layout.addWidget(self.add_button_button, 6, 0, 1, 3)
        layout.addWidget(button_area, 10, 0, 1, 3)

        for name, data in group_data.items():
            self.add_button(name, data)

    def add_button(self, name="", data=None):
        n = len(self.buttons["labels"])
        label = QLabel("text:")
        lineEdit = QLineEdit(name)
        button = QPushButton("remove button")
        button.clicked.connect(lambda state=None, x=n: self.remove_button(x))
        table = Channels_Check_Table(
            self,
            ["set", "channel", "value"],
            only_output=True,
            info_dict=data,
            checkstrings=[2],
        )
        line = QFrame()
        line.setFrameShadow(QFrame.Raised)
        line.setLineWidth(5)
        line.setFrameShape(QFrame.HLine)
        self.button_widge.layout().addWidget(label, n * 3, 0)
        self.button_widge.layout().addWidget(lineEdit, n * 3, 1)
        self.button_widge.layout().addWidget(button, n * 3, 2)
        self.button_widge.layout().addWidget(table, 1 + n * 3, 0, 1, 3)
        self.button_widge.layout().addWidget(line, 2 + n * 3, 0, 1, 3)
        self.buttons["labels"].append(label)
        self.buttons["texts"].append(lineEdit)
        self.buttons["tables"].append(table)
        self.buttons["removers"].append(button)
        self.buttons["lines"].append(line)

    def remove_button(self, n):
        for i, label in enumerate(self.buttons["labels"]):
            if i < n:
                continue
            lineEdit = self.buttons["texts"][i]
            button = self.buttons["removers"][i]
            table = self.buttons["tables"][i]
            line = self.buttons["lines"][i]
            button.clicked.disconnect()
            if i > n:
                self.button_widge.layout().removeWidget(label)
                self.button_widge.layout().removeWidget(lineEdit)
                self.button_widge.layout().removeWidget(button)
                self.button_widge.layout().removeWidget(table)
                self.button_widge.layout().removeWidget(line)
                self.button_widge.layout().addWidget(label, (i - 1) * 3, 0)
                self.button_widge.layout().addWidget(lineEdit, (i - 1) * 3, 1)
                self.button_widge.layout().addWidget(button, (i - 1) * 3, 2)
                self.button_widge.layout().addWidget(table, 1 + (i - 1) * 3, 0, 1, 3)
                self.button_widge.layout().addWidget(line, 2 + (i - 1) * 3, 0, 1, 3)
                button.clicked.connect(
                    lambda state=None, x=i - 1: self.remove_button(x)
                )
            else:
                self.button_widge.layout().removeWidget(label)
                self.button_widge.layout().removeWidget(lineEdit)
                self.button_widge.layout().removeWidget(button)
                self.button_widge.layout().removeWidget(table)
                self.button_widge.layout().removeWidget(line)
        for v in self.buttons.values():
            widget = v.pop(n)
            widget.deleteLater()

    def get_data(self):
        button_data = {}
        for i, line in enumerate(self.buttons["texts"]):
            name = line.text()
            if not name or name in button_data:
                raise Exception(
                    f"Buttons must be named and two buttons should not be named the same!\n{name}"
                )
            button_data[name] = self.buttons["tables"][i].get_info()
        name = self.lineEdit_group_name.text()
        if not name:
            raise Exception("Button groups must be named!")
        return {name: button_data}


if __name__ == "__main__":
    app = QApplication()
    panel = Set_Panel_Config()
    panel.show()
    import sys

    sys.exit(app.exec())



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'read_channels.ui'
##
## Created by: Qt User Interface Compiler version 6.5.0
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (
    QCoreApplication,
    QDate,
    QDateTime,
    QLocale,
    QMetaObject,
    QObject,
    QPoint,
    QRect,
    QSize,
    QTime,
    QUrl,
    Qt,
)
from PySide6.QtGui import (
    QBrush,
    QColor,
    QConicalGradient,
    QCursor,
    QFont,
    QFontDatabase,
    QGradient,
    QIcon,
    QImage,
    QKeySequence,
    QLinearGradient,
    QPainter,
    QPalette,
    QPixmap,
    QRadialGradient,
    QTransform,
)
from PySide6.QtWidgets import QApplication, QCheckBox, QGridLayout, QSizePolicy, QWidget


class Ui_read_channels_config(object):
    def setupUi(self, read_channels_config):
        if not read_channels_config.objectName():
            read_channels_config.setObjectName("read_channels_config")
        read_channels_config.resize(372, 230)
        self.gridLayout = QGridLayout(read_channels_config)
        self.gridLayout.setObjectName("gridLayout")
        self.gridLayout.setContentsMargins(0, 0, 0, 0)
        self.checkBox_read_all = QCheckBox(read_channels_config)
        self.checkBox_read_all.setObjectName("checkBox_read_all")

        self.gridLayout.addWidget(self.checkBox_read_all, 0, 0, 1, 1)

        self.checkBox_split_trigger = QCheckBox(read_channels_config)
        self.checkBox_split_trigger.setObjectName("checkBox_split_trigger")

        self.gridLayout.addWidget(self.checkBox_split_trigger, 0, 1, 1, 1)

        self.checkBox_read_variables = QCheckBox(read_channels_config)
        self.checkBox_read_variables.setObjectName("checkBox_read_variables")

        self.gridLayout.addWidget(self.checkBox_read_variables, 0, 2, 1, 1)

        self.retranslateUi(read_channels_config)

        QMetaObject.connectSlotsByName(read_channels_config)

    # setupUi

    def retranslateUi(self, read_channels_config):
        read_channels_config.setWindowTitle(
            QCoreApplication.translate("read_channels_config", "Form", None)
        )
        self.checkBox_read_all.setText(
            QCoreApplication.translate("read_channels_config", "Read All", None)
        )
        self.checkBox_split_trigger.setText(
            QCoreApplication.translate(
                "read_channels_config", "Split trigger and read", None
            )
        )
        self.checkBox_read_variables.setText(
            QCoreApplication.translate(
                "read_channels_config", "read/save variables", None
            )
        )

    # retranslateUi



def getItemIndex(item_model, data: str, starts_with=False):
    """Iteration over the item_model to return the index of an item with
    given data.

    Parameters
    ----------
    item_model : QStandardItemModel
        the searched item_model
    data : str
        the data which the looked-for item has
    starts_with : bool, default False
        also return an item, if it only starts with the searched data (Default value = False)
    data:str :


    Returns
    -------


    """
    rows = item_model.rowCount()
    cols = item_model.columnCount()
    for r in range(rows):
        for c in range(cols):
            item = item_model.item(r, c)
            if item is None:
                continue
            tester = (
                item.data().startswith(data) if starts_with else item.data() == data
            )
            if tester:
                return item.index()
            if item.hasChildren():
                found, ind = iterItem(item, data, starts_with)
                if found:
                    return ind
    return None


def iterItem(item, data, startWith=False):
    """Iteration over the children of the given item to return the index
    of an item with given data. Called by getItemIndex.
    Runs recursively, if the children-items also have children.

    Parameters
    ----------
    item : QStandardItem
        item whose children are iterated over
    data : str
        the data which the looked-for item has
    startWith : bool, default False
        also return an item, if it only starts with the searched data (Default value = False)

    Returns
    -------


    """
    rows = item.rowCount()
    cols = item.columnCount()
    found = False
    ind = None
    for r in range(rows):
        if found:
            break
        for c in range(cols):
            if found:
                break
            child = item.child(r, c)
            if child is None:
                continue
            tester = (
                child.data().startswith(data) if startWith else child.data() == data
            )
            if tester:
                return True, child.index()
            if child.hasChildren():
                found, ind = iterItem(child, data)
    return found, ind


def get_substeps(parent):
    """Helper function, to get the substeps of loop_step with children.

    Parameters
    ----------
    parent :


    Returns
    -------

    """
    steps = []
    for i in range(parent.rowCount()):
        item = parent.child(i, 0)
        substeps = get_substeps(item)
        steps.append((item.data(), substeps))
    return steps



from nomad_camels.main_classes.loop_step import Loop_Step, Loop_Step_Config
from nomad_camels.ui_widgets.channels_check_table import Call_Functions_Table
from nomad_camels.utility import variables_handling


class Call_Function(Loop_Step):
    def __init__(self, name="", parent_step=None, step_info=None, **kwargs):
        super().__init__(name, parent_step, step_info, **kwargs)
        self.step_type = "Call Function"
        if step_info is None:
            step_info = {}
        self.call_functions = (
            step_info["call_functions"] if "call_functions" in step_info else []
        )
        self.update_used_devices()

    def update_used_devices(self):
        self.used_devices = []
        for func in self.call_functions:
            device = func.split(".")[0]
            if device not in self.used_devices:
                self.used_devices.append(device)

    def get_protocol_string(self, n_tabs=1):
        tabs = "\t" * n_tabs
        protocol_string = super().get_protocol_string(n_tabs)
        available_functions = variables_handling.get_non_channel_functions()
        for func in self.call_functions:
            if func not in available_functions:
                raise Exception(
                    f'Trying to call function "{func}" but it is not available!\nCheck whether you maybe renamed an instrument.'
                )
            dev, fun = func.split(".")
            protocol_string += f'{tabs}devs["{dev}"].{fun}()\n'
        return protocol_string

    def get_protocol_short_string(self, n_tabs=0):
        short_string = super().get_protocol_short_string(n_tabs)
        short_string = f"{short_string[:-1]} - {self.call_functions}\n"
        return short_string


class Call_Function_Config(Loop_Step_Config):
    def __init__(self, loop_step: Call_Function, parent=None):
        super().__init__(parent, loop_step)
        labels = ["call?", "function"]
        info = {"functions": self.loop_step.call_functions}
        self.table = Call_Functions_Table(
            headerLabels=labels, parent=self, info_dict=info, title="Functions"
        )
        self.layout().addWidget(self.table, 2, 0, 1, 5)

    def update_step_config(self):
        super().update_step_config()
        info = self.table.get_info()
        self.loop_step.call_functions = info["functions"]



from PySide6.QtWidgets import (
    QWidget,
    QLabel,
    QGridLayout,
    QLineEdit,
    QTextEdit,
    QComboBox,
    QCheckBox,
)

from nomad_camels.main_classes.loop_step import Loop_Step, Loop_Step_Config


class Prompt_Loop_Step(Loop_Step):
    """
    This step displays a prompt (QMessageBox) at runtime and pauses the
    execution of the protocol until the user clicks "ok" in the prompt. This may
    be used for example to prompt the user to do something at the setup.

    Attributes
    ----------
    short_test : str
        This will be the window title of the prompt.
    long_test : str
        This will be the full text, displayed inside the window.
    icon : str
        If 'Error', the `QMessagebox.Critical` icon is displayed, if 'Warning',
        then `QMessagebox.Warning` is used, otherwise `QMessagebox.Information`.
    """

    def __init__(self, name="", parent_step=None, step_info=None, **kwargs):
        super().__init__(name, parent_step, step_info, **kwargs)
        self.step_type = "Prompt"
        if step_info is None:
            step_info = {}
        self.short_text = step_info["short_text"] if "short_text" in step_info else ""
        self.long_text = step_info["long_text"] if "long_text" in step_info else ""
        self.icon = step_info["icon"] if "icon" in step_info else "Info"
        self.abortable = step_info["abortable"] if "abortable" in step_info else False

    def get_protocol_string(self, n_tabs=1):
        """Sets the prompt's `done_flag` to False, then starts execution of the
        prompt and waits until it is done_flag."""
        tabs = "\t" * n_tabs
        protocol_string = super().get_protocol_string(n_tabs)
        protocol_string += f'{tabs}boxes["prompt_{self.name}"].done_flag = False\n'
        protocol_string += f'{tabs}boxes["prompt_{self.name}"].helper.executor.emit()\n'
        protocol_string += f'{tabs}while not boxes["prompt_{self.name}"].done_flag:\n'
        protocol_string += f"{tabs}\tyield from bps.sleep(0.1)\n"
        if self.abortable:
            protocol_string += f'{tabs}\tif boxes["prompt_{self.name}"].abort_flag:\n'
            protocol_string += f"{tabs}\t\treturn\n"
        return protocol_string

    def get_add_main_string(self):
        """Adds the setup of the box to the `steps_add_main` function of the
        protocol."""
        long_text = self.long_text.replace("\n", "\\n").replace('"', '\\"')
        short_text = self.short_text.replace('"', '\\"').replace("\n", "\\n")
        add_main_string = super().get_add_main_string()
        add_main_string += f'\tboxes["prompt_{self.name}"] = helper_functions.Prompt_Box("{self.icon}", "{long_text}", "{short_text}", abortable={self.abortable})\n'
        return add_main_string


class Prompt_Loop_Step_Config(Loop_Step_Config):
    """ """

    def __init__(self, loop_step: Prompt_Loop_Step, parent=None):
        super().__init__(parent, loop_step)
        self.sub_widget = Prompt_Loop_Step_Config_Sub(loop_step, self)
        self.layout().addWidget(self.sub_widget, 1, 0, 1, 5)


class Prompt_Loop_Step_Config_Sub(QWidget):
    """The QLineEdit and labels to make everything clear are provided."""

    def __init__(self, loop_step: Prompt_Loop_Step, parent=None):
        super().__init__(parent)
        self.loop_step = loop_step

        label1 = QLabel("Title:")
        label2 = QLabel("Text:")
        label3 = QLabel("Icon:")
        self.lineEdit_short = QLineEdit(self)
        self.lineEdit_short.setText(loop_step.short_text)
        self.lineEdit_short.textChanged.connect(self.update_info)

        self.textEdit_long = QTextEdit(self)
        self.textEdit_long.setText(loop_step.long_text)
        self.textEdit_long.textChanged.connect(self.update_info)

        self.comboBox_icon = QComboBox()
        icons = ["Information", "Warning", "Critical"]
        self.comboBox_icon.addItems(icons)
        self.comboBox_icon.setCurrentText(loop_step.icon)
        self.comboBox_icon.currentTextChanged.connect(self.update_info)

        self.checkBox_abortable = QCheckBox("can abort protocol")
        self.checkBox_abortable.setChecked(loop_step.abortable)
        self.checkBox_abortable.stateChanged.connect(self.update_info)

        layout = QGridLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(label3, 0, 0)
        layout.addWidget(self.comboBox_icon, 0, 1)
        layout.addWidget(self.checkBox_abortable, 1, 0, 1, 2)
        layout.addWidget(label1, 2, 0)
        layout.addWidget(self.lineEdit_short, 2, 1)
        layout.addWidget(label2, 3, 0)
        layout.addWidget(self.textEdit_long, 3, 1)
        self.setLayout(layout)

    def update_info(self):
        """ """
        self.loop_step.short_text = self.lineEdit_short.text()
        self.loop_step.long_text = self.textEdit_long.toPlainText()
        self.loop_step.icon = self.comboBox_icon.currentText()
        self.loop_step.abortable = self.checkBox_abortable.isChecked()



from PySide6.QtWidgets import QWidget

from nomad_camels.main_classes.loop_step import Loop_Step, Loop_Step_Config

from nomad_camels.gui.gradient_descent_step import Ui_Grad_Desc
from nomad_camels.ui_widgets.add_remove_table import AddRemoveTable
from nomad_camels.utility import variables_handling
from nomad_camels.bluesky_handling import builder_helper_functions
from nomad_camels.frontpanels.plot_definer import Plot_Info


class Gradient_Descent_Step(Loop_Step):
    """
    A step performing a gradient descent with a given channel to optimize a
    given function.


    Attributes
    ----------
    read_channels : list[str]
        A list of all the channels which are read for the optimization.
    extremum : str ('Minimum', 'Maximum')
        The extremum type that should be found.
    out_channel : str
        The channel which is used for the optimization.
    opt_func : str
        This string is evaluated by the given to give the target function.
    start_val : str, float
        Representation of where to start the algorithm.
    min_val : str, float
        The minimum value that should be given to the `set_channel`.
    max_val : str, float
        The maximum value that should be given to the `set_channel`.
    learning_rate : str, float
         A weight for the learning of the gradient descent.
         The next shift `delta_w` is calculated as:
         delta_w = -learning_rate * <current_gradient> + momentum * <last_delta_w>
    threshold : str, float
        If the difference between two measurements is smaller than this
        threshold, the algorithm recognizes the value as the optimum and stops.
    momentum : str, float
         (Default value = 0.8)
         A momentum to keep up the last direction.
         The next shift `delta_w` is calculated as:
         delta_w = -learning_rate * <current_gradient> + momentum * <last_delta_w>
    min_step : str, float
        The minimum step size.
    max_step : str, float
        The maximum step size.
    n_steps : str, int
        The maximum number of iterations until the algorithm should stop if it
        did not arrive at the threshold yet.
    plot_steps : bool
        Whether to plot the single steps of the algorithm at runtime.
    """

    def __init__(self, name="", parent_step=None, step_info=None, **kwargs):
        super().__init__(name, parent_step, step_info, **kwargs)
        self.step_type = "Gradient Descent"
        step_info = step_info or {}
        self.read_channels = (
            step_info["read_channels"] if "read_channels" in step_info else []
        )
        self.extremum = step_info["extremum"] if "extremum" in step_info else "Minimum"
        self.out_channel = (
            step_info["out_channel"] if "out_channel" in step_info else ""
        )
        self.opt_func = step_info["opt_func"] if "opt_func" in step_info else ""
        self.start_val = step_info["start_val"] if "start_val" in step_info else ""
        self.min_val = step_info["min_val"] if "min_val" in step_info else ""
        self.max_val = step_info["max_val"] if "max_val" in step_info else ""
        self.learning_rate = (
            step_info["learning_rate"] if "learning_rate" in step_info else ""
        )
        self.threshold = step_info["threshold"] if "threshold" in step_info else ""
        self.momentum = step_info["momentum"] if "momentum" in step_info else ""
        self.min_step = step_info["min_step"] if "min_step" in step_info else ""
        self.max_step = step_info["max_step"] if "max_step" in step_info else ""
        self.n_steps = step_info["n_steps"] if "n_steps" in step_info else ""
        self.plot_steps = step_info["plot_steps"] if "plot_steps" in step_info else True

    def update_used_devices(self):
        """Uses all devices in self.read_channels and of self.out_channel"""
        self.used_devices = []
        set_device = variables_handling.channels[self.out_channel].device
        self.used_devices.append(set_device)
        for channel in self.read_channels:
            if channel in variables_handling.channels:
                device = variables_handling.channels[channel].device
                if device not in self.used_devices:
                    self.used_devices.append(device)

    def get_outer_string(self):
        """Adds the plot, if self.plot_steps is True. The plot displays the
        formula to be optimized vs the out_channel."""
        if self.plot_steps:
            plot = Plot_Info(
                x_axis=self.out_channel,
                y_axes={"formula": [self.opt_func], "axis": ["left"]},
                title="gradient-descent",
            )
            return builder_helper_functions.plot_creator(
                [plot], f"create_plots_{self.name}"
            )[0]
        return ""

    def get_add_main_string(self):
        """Adds the call of creating the plots if self.plot_steps."""
        add_main_string = ""
        if self.plot_steps:
            stream = f'"{self.name}"'
            add_main_string += builder_helper_functions.get_plot_add_string(
                self.name, stream
            )
        return add_main_string

    def get_protocol_string(self, n_tabs=1):
        """Evaluates all / most of the values with the protocol's evaluator and
        calls the helper function gradient_descent to perform the algorithm."""
        tabs = "\t" * n_tabs
        func_text = f'"{self.opt_func}"'
        if self.extremum == "Maximum":
            func_text = f'"-({self.opt_func})"'

        protocol_string = super().get_protocol_string(n_tabs)
        protocol_string += f"{tabs}channels = ["
        for i, channel in enumerate(self.read_channels):
            if channel not in variables_handling.channels:
                raise Exception(
                    f"Trying to read channel {channel} in {self.full_name}, but it does not exist!"
                )
            if i > 0:
                protocol_string += ", "
            name = variables_handling.channels[channel].name
            if "." in name:
                dev, chan = name.split(".")
                protocol_string += f'devs["{dev}"].{chan}'
            else:
                protocol_string += f'devs["{name}"]'
        protocol_string += "]\n"

        name = variables_handling.channels[self.out_channel].name
        if "." in name:
            dev, chan = name.split(".")
            setter = f'devs["{dev}"].{chan}'
        else:
            setter = f'devs["{name}"]'

        protocol_string += f'{tabs}yield from helper_functions.gradient_descent(eva.eval("{self.n_steps}"), eva.eval("{self.threshold}"), eva.eval("{self.start_val}"), {func_text}, eva, {setter}, channels, eva.eval("{self.min_step}"), eva.eval("{self.max_step}"), eva.eval("{self.min_val}"), eva.eval("{self.max_val}"), "{self.name}", eva.eval("{self.learning_rate}"), eva.eval("{self.momentum}"))\n'
        return protocol_string

    def get_protocol_short_string(self, n_tabs=0):
        """Describes the kind of extremum, the opt_func and the out_channel."""
        short_string = super().get_protocol_short_string(n_tabs)[:-1]
        short_string += (
            f" - {self.extremum} of {self.opt_func} via {self.out_channel}\n"
        )
        return short_string


class Gradient_Descent_Config(Loop_Step_Config):
    """ """

    def __init__(self, loop_step: Gradient_Descent_Step, parent=None):
        super().__init__(parent, loop_step)
        self.sub_widget = Gradient_Descent_Config_Sub(loop_step, self)
        self.layout().addWidget(self.sub_widget, 1, 0, 1, 5)

    def update_step_config(self):
        """ """
        super().update_step_config()
        self.sub_widget.update_step_config()


class Gradient_Descent_Config_Sub(Ui_Grad_Desc, QWidget):
    """ """

    def __init__(self, loop_step: Gradient_Descent_Step, parent=None):
        super().__init__(parent)
        self.setupUi(self)
        self.loop_step = loop_step
        self.read_table = AddRemoveTable(
            title="Read Channels",
            headerLabels=[],
            tableData=loop_step.read_channels,
            comboBoxes=variables_handling.channels.keys(),
        )
        self.layout().addWidget(self.read_table, 20, 0, 1, 3)
        self.comboBox_extremum_type.addItems(["Minimum", "Maximum"])
        self.comboBox_extremum_type.setCurrentText(loop_step.extremum)

        out_box = []
        for channel in variables_handling.channels:
            if variables_handling.channels[channel].output:
                out_box.append(channel)
        self.comboBox_output_channel.addItems(out_box)
        if loop_step.out_channel in out_box:
            self.comboBox_output_channel.setCurrentText(loop_step.out_channel)
        self.lineEdit_opt_func.setText(loop_step.opt_func)
        self.lineEdit_starting_val.setText(loop_step.start_val)
        self.lineEdit_min_val.setText(loop_step.min_val)
        self.lineEdit_max_val.setText(loop_step.max_val)
        self.lineEdit_learning_rate.setText(loop_step.learning_rate)
        self.lineEdit_threshold.setText(loop_step.threshold)
        self.lineEdit_momentum.setText(loop_step.momentum)
        self.lineEdit_smallest_step.setText(loop_step.min_step)
        self.lineEdit_largest_step.setText(loop_step.max_step)
        self.lineEdit_max_n_steps.setText(loop_step.n_steps)
        self.checkBox_plot_steps.setChecked(loop_step.plot_steps)

    def update_step_config(self):
        """ """
        self.loop_step.extremum = self.comboBox_extremum_type.currentText()
        self.loop_step.out_channel = self.comboBox_output_channel.currentText()
        self.loop_step.opt_func = self.lineEdit_opt_func.text()
        self.loop_step.start_val = self.lineEdit_starting_val.text()
        self.loop_step.min_val = self.lineEdit_min_val.text()
        self.loop_step.max_val = self.lineEdit_max_val.text()
        self.loop_step.learning_rate = self.lineEdit_learning_rate.text()
        self.loop_step.threshold = self.lineEdit_threshold.text()
        self.loop_step.momentum = self.lineEdit_momentum.text()
        self.loop_step.min_step = self.lineEdit_smallest_step.text()
        self.loop_step.max_step = self.lineEdit_largest_step.text()
        self.loop_step.n_steps = self.lineEdit_max_n_steps.text()
        self.loop_step.read_channels = self.read_table.update_table_data()
        self.loop_step.plot_steps = self.checkBox_plot_steps.isChecked()



import pytest
from PySide6.QtWidgets import QApplication, QMainWindow, QColorDialog
from PySide6.QtGui import QCloseEvent, QColor
from unittest.mock import patch
from unittest.mock import MagicMock
import pyqtgraph as pg


# Importing the necessary components for the test
from nomad_camels.main_classes.plot_pyqtgraph import PlotWidget, LivePlot, Plot_Options


@pytest.fixture(scope="module")
def qapp():
    """Fixture for creating a QApplication instance"""
    app = QApplication.instance()
    if app is None:
        app = QApplication([])
    return app


@pytest.fixture
def plot_widget(qapp):
    """Fixture for creating an instance of PlotWidget."""
    return PlotWidget(
        x_name="time",
        y_names=["value1", "value2"],
        stream_name="primary",
        use_bluesky=False,  # Mocking bluesky usage for simplicity
    )


def test_initialization(plot_widget):
    """Test the initialization of PlotWidget."""
    assert plot_widget.x_name == "time"
    assert plot_widget.y_names == ["value1", "value2"]
    assert plot_widget.stream_name == "primary"
    assert plot_widget.plot_widget is not None
    assert plot_widget.livePlot is not None


def test_show_again(plot_widget):
    """Test the show_again method."""
    with patch.object(plot_widget, "show") as mock_show:
        plot_widget.show_again()
        mock_show.assert_not_called()
        plot_widget.close()
        plot_widget.show_again()
        mock_show.assert_called_once()


def test_auto_range(plot_widget):
    """Test the auto_range method."""
    with patch.object(plot_widget.livePlot.plotItem.vb, "autoRange") as mock_autoRange:
        plot_widget.auto_range()
        mock_autoRange.assert_called_once()
        if plot_widget.ax2_viewbox:
            with patch.object(
                plot_widget.ax2_viewbox, "autoRange"
            ) as mock_ax2_autoRange:
                plot_widget.auto_range()
                mock_ax2_autoRange.assert_called_once()


def test_change_maxlen(plot_widget):
    """Test the change_maxlen method."""
    plot_widget.lineEdit_n_data.setText("100")
    plot_widget.change_maxlen()
    assert plot_widget.livePlot.maxlen == 100

    plot_widget.lineEdit_n_data.setText("inf")
    plot_widget.change_maxlen()
    assert plot_widget.livePlot.maxlen == float("inf")

    plot_widget.lineEdit_n_data.setText("invalid")
    plot_widget.change_maxlen()
    # maxlen should not change due to invalid input
    assert plot_widget.livePlot.maxlen == float("inf")


def test_show_options(plot_widget):
    """Test the show_options method."""
    plot_widget.show_options()
    assert plot_widget.options_open
    assert plot_widget.pushButton_show_options.text() == "Hide Options"

    plot_widget.show_options()
    assert not plot_widget.options_open
    assert plot_widget.pushButton_show_options.text() == "Show Options"


def test_clear_plot(plot_widget):
    """Test the clear_plot method."""
    with patch.object(plot_widget.livePlot, "clear_plot") as mock_clear_plot:
        for fit_plot in plot_widget.liveFitPlots:
            with patch.object(fit_plot, "clear_plot") as mock_fit_clear_plot:
                plot_widget.clear_plot()
                mock_clear_plot.assert_called_once()
                mock_fit_clear_plot.assert_called_once()


def test_closeEvent(plot_widget):
    """Test the closeEvent method."""
    mock_slot = MagicMock()
    plot_widget.closing.connect(mock_slot)
    plot_widget.closeEvent(QCloseEvent())
    mock_slot.assert_called_once()


@pytest.fixture
def main_window(qapp):
    window = QMainWindow()
    return window


@pytest.fixture
def live_plot(main_window):
    plot_widget = pg.PlotWidget()
    plot_item = plot_widget.getPlotItem()
    plot_item.setLogMode = MagicMock()
    ax2_viewbox = pg.ViewBox()
    plot_item.scene().addItem(ax2_viewbox)
    live_plot = LivePlot("time", ["test"], plot_item=plot_item, ax2_viewbox=ax2_viewbox)
    return live_plot


@pytest.fixture
def plot_options(main_window, live_plot):
    plot_options = Plot_Options(parent=main_window, livePlot=live_plot)
    return plot_options
